#include "fraise.h"
#define _NUM_PPFRAMES 8

#define TRUE 1
#define FALSE 0
using namespace std;

CvMemStorage *__mmstg;

int main (int argc, const char **argv)
{
  int _frame_count = 0, nframes, _end_reached = FALSE, algo;
  double _time_taken, _total_exec_time, _start_time;
  double scale;
 
  vector<string> list;
  // For storing each images into a vector that could be used in future
  vector<IplImage*> frames;
 
  IplImage *imgsrc = NULL, *imgdst = NULL;
  // H homography between consecutive frames;
  // _psurfH : Homography between frames apart by _NUM_PPFRAMES
  CvMat *H = NULL, *_psurfH = NULL;
  //SURF is used as an error correction measure
  CvSURFParams params = cvSURFParams(500, 1);
  CvSeq *srckpts = NULL, *srcdes =NULL;
  CvSeq *dstkpts = NULL, *dstdes =NULL;
  int i, j;
  char *fname = (char *)malloc(MAX_PATH_LEN * sizeof(char));
#ifdef HAVE_VIDEO
  /* Sequence to store cumulative homography elements */
  CvSeq *xseq;
  CvMat *h = NULL;
  /* Video Input and output structures */
  CvCapture* vfile = NULL;
	
  IplImage *grayframe = NULL, *colorframe = NULL;
  double fps, elem;
  int _frame_width, _frame_height, _elem_count;
  char *pref = (char *)malloc(MAX_PATH_LEN * sizeof(char));
  CvMat  **vH;
#else
  CvMat **xmat = NULL;
#endif
  /* Argument checking block */
#ifdef HAVE_VIDEO
  if (argc != 5) {
    fprintf(stderr, "Usage: %s\n" 
            "\t</path/to/video/filename> \n"
            "\t<resolution_factor between 1-2> \n"
            "\t</path/to/output/dir>\n" 
            "\t<algo to use:0-3>\n", argv[0]);
    return -1;
  }
  algo = atoi(argv[4]);	
#else
  if (argc != 6) {
    fprintf(stderr, "Insufficient arguments.\nUsage: %s <inputdir> <extension> <resize> <output dir> <algo>\n", 
            argv[0]);
    return -1;
  }
	
  algo = atoi(argv[5]);	
#endif
  /* Enabling SIFT support, requires 3rd part software */
#ifndef HAVESIFT
  if (algo==1) {
    fprintf (stderr, "ERR: Please compile with -DHAVESIFT to enable this option\n");
    return -1;
  } 
#endif
  if (algo<0 || algo>3) {
    fprintf (stderr, "ERR: The algorithm switch should be an integer between 0-3, inclusive\n");
    return -1;
  } 
  //used by video related code and surf based matching
  __mmstg = cvCreateMemStorage (0);
  // Initialize Homography Sequence
  H = cvCreateMat (3, 3, CV_32FC1);
  _psurfH = cvCloneMat (H);
  cvSetIdentity (_psurfH, cvRealScalar(1));

#ifdef HAVE_VIDEO
  vfile = cvCreateFileCapture (argv[1]);
  if (!vfile) {
    fprintf (stderr, "ERR:Cannot initialize capture. Aborting\n");
    return -1;
  }
  /* Initializing the input for further processing */
  scale = strtod(argv[2], NULL);
  /* Retrieving information from video file */
  fps = cvGetCaptureProperty (vfile, CV_CAP_PROP_FPS);
  cvGrabFrame(vfile);
  _frame_width = (int)cvGetCaptureProperty(vfile, CV_CAP_PROP_FRAME_WIDTH);
  _frame_height = (int)cvGetCaptureProperty(vfile, CV_CAP_PROP_FRAME_HEIGHT);
  fprintf(stdout, "INF:Input Video Information:\nFrames per sec: %2.3f\nResolution: %dx%d\n",
                   fps, _frame_width, _frame_height);
  grayframe = cvCreateImage (cvSize(_frame_width, _frame_height), 8, 1);
  /* Is used to store cumulative homography temporarily */
  h = cvCreateMat (3, 3, CV_32FC1);
  cvSetIdentity (H, cvRealScalar(1));
  
  xseq = cvCreateSeq(0, sizeof(CvSeq), sizeof(double), __mmstg);
  for (i =0; i <9; i++) {
    elem = cvmGet (H, i/3, i%3);
    cvSeqPush (xseq, &elem);
  }	
  _check_directories(argv[3]);
#else
	/* The number of frames are known in advance */
  getFilelist(list, string(argv[1]), string(argv[2]));
  nframes = list.size();
  if (!nframes) {
    fprintf (stderr, "ERR:No frames in the directory %s with extension %s. Aborting\n",
            argv[1], argv[2]);
    return -1;
  }
  scale = strtod(argv[3], NULL);
  xmat = (CvMat **)cvAlloc (nframes* sizeof(CvMat));
  xmat[0] = cvCreateMat (3, 3, CV_32FC1);
  cvSetIdentity (xmat[0], cvRealScalar(1));
  _check_directories(argv[4]);
#endif
		
  /* The Registration logic starts from this block */
  cvUseOptimized(1);
  cvSetNumThreads(0);
  _start_time = (double) cvGetTickCount ();
  while (!_end_reached) {
    if (_frame_count==0){
#ifdef HAVE_VIDEO
      /* Load reference frame */
      colorframe = cvQueryFrame (vfile);
      cvCvtColor (colorframe, grayframe, CV_BGR2GRAY);
      imgsrc = cvCreateImage (cvSize((int)(_frame_width*scale), (int)(_frame_height*scale)),
                IPL_DEPTH_8U, 1);
      cvResize (grayframe, imgsrc, CV_INTER_LINEAR);
      /* Reset _elem_count for future use */
      _elem_count = 0;
#else
      imgsrc = loadReferenceImage (argv[1], list[_frame_count].c_str(), 1, scale);
#endif
	  //cvEqualizeHist (imgsrc, imgsrc);
      /* Vector of gray scale frames to display alignment results */
      frames.push_back(cvCloneImage(imgsrc));
    }

#ifdef HAVE_VIDEO
    if (colorframe == cvQueryFrame (vfile)) {
      cvCvtColor (colorframe, grayframe, CV_BGR2GRAY);
      imgdst = cvCreateImage (cvSize((int)(_frame_width*scale), (int)(_frame_height*scale)),
                 IPL_DEPTH_8U, 1);
      cvResize (grayframe, imgdst, CV_INTER_LINEAR);
#else
      /*Since we are computing H between ith and (i+1)th frames */
      if (_frame_count < nframes - 1) {
        imgdst = loadReferenceImage (argv[1], list[_frame_count+1].c_str(), 1, scale);
#endif
        _time_taken = (double) cvGetTickCount ();
        /* Compute Homography using any of the methods */
        getHomography (imgsrc, imgdst, &H, algo);
        fprintf (stdout, "INF:Computing H b/w frames %04d/%04d in %4.5fms\n",
          _frame_count, _frame_count+1, 
          ((double)cvGetTickCount () - _time_taken)/(cvGetTickFrequency()*1000.));
         //displayElementsWI (H, "SOF H");
#ifdef HAVE_VIDEO
        /* Set temporary matrix with values from sequence */
        for (i =0; i <9; i++)
          cvmSet(h, i/3, i%3, *((double *)cvGetSeqElem (xseq, _elem_count++)));
          /* h would contain the cumulative Homography */
          cvMatMul(H, h, h);
          /* Store elements of cumumalative Homography matrix in sequence seq */
          for (i =0; i <9; i++) {
            elem = cvmGet (h, i/3, i%3);
            cvSeqPush (xseq, &elem);
          }
#else
          /* Calculate Cumulative Homography */
          xmat[_frame_count+1] = cvCreateMat (3, 3, CV_32FC1);
          cvMatMul(H, xmat[_frame_count], xmat[_frame_count+1]);
#endif
        /* Reset source image to current image */
        cvCopy (imgdst, imgsrc, NULL);
        /* Vector of gray scale frames to display alignment results */
        frames.push_back(cvCloneImage(imgsrc));
        cvReleaseImage (&imgdst); 
      } else {
       _end_reached = TRUE;
      }
      _frame_count++;
   } 			
  /* Generating output */
  _total_exec_time = ((double)cvGetTickCount () - _start_time)/(cvGetTickFrequency()*1000.);
  fprintf (stdout, "========================================\n");
  fprintf (stdout, "INF:Execution summary: %d frames in %8.2fms\nINF:Average speed %4.2f frames/sec\n",
                    _frame_count, _total_exec_time, (double)_frame_count/(_total_exec_time/1000.));
  fprintf (stdout, "========================================\n");

  bzero(fname, MAX_PATH_LEN);

#ifdef HAVE_VIDEO
  nframes = _frame_count;
  sprintf (fname, "%s%s%s", argv[3], _PATH_SEP, XCOUNTS);

  vH = (CvMat **)cvAlloc (nframes* sizeof (CvMat));
  outputRegParams (xseq, vH, _frame_count, fname); 
	saveAlignedFrames (frames, vH, argv[3]);
  // The above command would be replaced by the following in future     
  //fprintf (stdout, "INF:aligned video generated in: %s\n", fname);
  //bzero(fname, MAX_PATH_LEN);
  //sprintf (fname, "%s%s%s", argv[3], _PATH_SEP, OUTPUTVIDEO);
  //generateVideo (frames, vH, fname);
    
  //TODO: Get global alignment and mosaic done
  //createMosaic(frames, vH, fname);
  for (_frame_count=0; _frame_count <nframes; _frame_count++)
    cvReleaseMat (&vH[_frame_count]);
  cvFree (vH);
#else
  fprintf (stdout, "INF:Saving Registered Images in: %s\n", argv[4]);
  sprintf (fname, "%s%s%s", argv[4], _PATH_SEP, XCOUNTS);
  outputRegParams (xmat, nframes, fname);
  saveAlignedFrames (frames, xmat, argv[4]);

  bzero(fname, MAX_PATH_LEN);
  sprintf (fname, "%s%s%s", argv[3], _PATH_SEP, MOSAIC);
  //createMosaic(frames, xmat, fname);

  for (_frame_count=0; _frame_count <nframes; _frame_count++)
    cvReleaseMat (&xmat[_frame_count]);
  cvFree (xmat);
#endif
			
  /* Cleaning up resources */
  free (fname);
  cvReleaseMat (&_psurfH);
  cvReleaseMat (&H);
  cvReleaseImage (&imgsrc);
  cvReleaseMemStorage (&__mmstg);

  return 0;
}