#include <cv.h>
#include <highgui.h>

#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <stdarg.h>
#include <libgen.h>
#include <dirent.h>
#include <math.h>
#include <time.h>
#include <sys/stat.h>
#include <sys/types.h>

#include <vector>
#include <deque>
#include <string>
#include <algorithm> 

#ifdef HAVESIFT
/* For SIFT specific routines */
#include "sift.h"
#include "imgfeatures.h"
#define H_SIFT 1
CvPoint2D32f *sift2pts (IplImage *imgsrc, int *count);
#endif

#define _MASKVAL_OUTLIER 0
#define _MASKVAL_INLIER 1
#define MAX_PATH_LEN 2048
#define _ALIGN_OUTP "Aligned"
#define XCOUNTS "Transcounts.txt"
#define MOSAIC "mosaic.jpg"
#define OUTPUTVIDEO "aligned.avi"

#ifdef WIN32
#define _PATH_SEP "\\"
#else
#define _PATH_SEP "/"
#endif

#ifndef ABS
#define ABS(x) ( ( x < 0 )? -x : x )
#endif

#ifndef bzero
#define bzero(b,len) (memset((b), '\0', (len)), (void) 0)
#endif

/* Number of features to use - increase for high resolution images */
#define MAX_COUNT 512
/* Features to chose from and then computing homogaphy */
#define H_KLTF 0
#define H_SURF 2
#define H_MSER 3
/* Fraction of points to consider from MSER output */
#define FRAC 0.0001
/* Number of frames after which F2F homography calculation is automatically 
 triggered */
#define TRIGGER_F2F_H 16
using namespace std;

extern CvMemStorage *__mmstg;
extern CvSeq *ptpairs;

int getHomography (IplImage *imgsrc, IplImage *imgdst, CvMat **H, int method);
/*functions to convert points from each of the extracted images to openCV 
 structure CvPoint2d32f for later use */
CvPoint2D32f *kltf2pts (IplImage *imgsrc, int *count);
CvPoint2D32f *surf2pts (IplImage *imgsrc, int *count);
CvPoint2D32f *mser2pts (IplImage *imgsrc, int *count);
/* Helper function for mser2pts */
CvMat *randidxs(int ub, double frac);
/* Functions for frame to frame homography estimation */
double compareSURFDescriptors (const float* d1, const float* d2, double best, int length);
int naiveNearestNeighbor ( const float* vec, int laplacian, const CvSeq* model_keypoints,
                          const CvSeq* model_descriptors );
void findHomographySURF (IplImage *imgsrc, IplImage *imgdst, CvMat **H);

/* IO specific function */
void _check_directories (const char *outputdir);
void getFilelist (vector<string>& list, string dir, string extn);
IplImage* loadReferenceImage (const char *dir, const char *file, const int conv2gray, const double rsz);
void createMosaic (vector<IplImage *>& images, CvMat **H, const char *fname);
void saveAlignedFrames (vector<IplImage *>& frames, CvMat **H, const char *outdir);
#ifdef HAVE_VIDEO
void outputRegParams(CvSeq *xseq, CvMat **vH, int nframes, const char *fname);
void generateVideo (vector<IplImage *>& frames, CvMat **H, const char *outfile);
#else
void outputRegParams (CvMat **xmat, int nframes, const char *fname);
#endif
void saveResult (IplImage *imgdst, CvMat *H, const char *origfname, const char *prefix);

void displayElementsWI (const CvMat *x, const char *marker);